Attribute VB_Name = "modHook"

'Dieser Source stammt von http://www.activevb.de
'und kann frei verwendet werden. Fr eventuelle Schden
'wird nicht gehaftet.
'
'Um Fehler oder Fragen zu klren, nutzen Sie bitte unser Forum.
'Ansonsten viel Spa und Erfolg mit diesem Source !
'
'****************************************************
'* modHook.bas                                      *
'* this module is the hooking backend of the class  *
'* Programmed: Konrad Rudolph (Mad Rat Pro.)        *
'* Last Change: 30.04.2003 21:23                    *
'* Version: 1.0.0                                   *
'****************************************************
Option Explicit

' this structure stores all relevant information about a hook
Public Type HookType
    Handle      As Long     ' handle to the parent object (ObjPtr)
    hWnd        As Long     ' handle to the subclassed window
    Messages()  As Long     ' array of hooked messages
    PrevProc    As Long     ' pointer to previous window procedure
End Type
'----------------------------------------------------

Private Declare Function CallWindowProc Lib "user32" Alias "CallWindowProcA" ( _
  ByVal lpPrevWndFunc As Long, _
  ByVal hWnd As Long, _
  ByVal Msg As Long, _
  ByVal wParam As Long, _
  ByVal lParam As Long _
) As Long
   
Private Declare Function SetWindowLong Lib "user32" Alias "SetWindowLongA" ( _
  ByVal hWnd As Long, _
  ByVal nIndex As Long, _
  ByVal dwNewLong As Long _
) As Long

Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" ( _
  nDest As Any, _
  nSrc As Any, _
  ByVal Length As Long _
)
'----------------------------------------------------

Private Hooks() As HookType     ' all hooking objects
Private bInit   As Boolean
'----------------------------------------------------

'----------------------------------------------------
' DoHook As HookType
' initializes the hooking structure for a new hook
' - IN ParentObj As Long: handle to the current CHook class
' - IN hWnd As Long: handle of the window to subclass
' - IN uMsg As Long: message to hook
' returns the filled in HookType structure
'----------------------------------------------------
Public Function DoHook( _
  ParentObj As Long, _
  hWnd As Long, _
  uMsg As Long _
) As HookType
    Dim PrevProc    As Long, _
        I           As Long
    
    ' redirect the message flow to our window procedure
    PrevProc = SetWindowLong( _
                            hWnd, _
                            (-4), _
                            AddressOf WndProc _
                        )
    
    If PrevProc <> 0 Then
        ' succeded...
        
        If Not bInit Then
            ' initialize the hooks array
            ReDim Preserve Hooks(10) As HookType
            bInit = True
        End If
        
        For I = 0 To UBound(Hooks)
            ' search first free hooking struct
            If Hooks(I).Handle = 0 Then _
                Exit For
        Next I
        
        ' no free slot found, enlarge array
        If I > UBound(Hooks) Then _
            ReDim Preserve Hooks(UBound(Hooks) + 10) As HookType
        
        ' assign data
        With DoHook
            .Handle = ParentObj
            .hWnd = hWnd
            ReDim Preserve .Messages(0) As Long
            .Messages(0) = uMsg
            .PrevProc = PrevProc
        End With
        
        ' store freshly created structure in local array
        Hooks(I) = DoHook
    End If
End Function

'----------------------------------------------------
' UnHook
' releases a subclassed window
' - IN Handle As Long: handle to the evoker class
'----------------------------------------------------
Public Sub UnHook(Handle As Long)
    Dim I As Long
    
    For I = 0 To UBound(Hooks)
        With Hooks(I)
            ' corresponding structure found
            If .Handle = Handle Then
                ' redirect window proc to default pointer
                Call SetWindowLong(.hWnd, (-4), .PrevProc)
                .Handle = 0
                Erase .Messages
            End If
        End With
    Next
End Sub

'----------------------------------------------------
' SyncHook As HookType
' syncronises the hook structure of a class with the one
' locally stored
' this procedure goes both directions:
' * to assign the classes struct to the module, write:
'   Call SyncHook(m_Hook)
' * to update the struct in the class with the data from
'   the module, simply write:
'   m_Hook = SyncHook(m_Hook)
'   Call SyncHook(m_Hook)    <-- this line's VERY important!
'
' - IN Hook AS HookType: the hooking structure from the class
' returns the hooking structure from the module
'----------------------------------------------------
Public Function SyncHook(Hook As HookType) As HookType
    Dim I As Long
    
    For I = 0 To UBound(Hooks)
        If Hooks(I).Handle = Hook.Handle Then
            ' 'swap' the structures:
            ' return the one from the module...
            SyncHook = Hooks(I)
            ' ...and assign the one from the class to the module
            Hooks(I) = Hook
        End If
    Next
End Function
'----------------------------------------------------

'----------------------------------------------------
' WndProc As Long
' our tiny window procedure that handles all incoming messages
' - IN hWnd As Long: look in the MSDN for details
' - IN uMsg As Long: look in the MSDN for details
' - IN wParam As Long: look in the MSDN for details
' - IN lParam As Long: look in the MSDN for details
' return value: look in the MSDN for details
'----------------------------------------------------
Public Function WndProc( _
  ByVal hWnd As Long, _
  ByVal uMsg As Long, _
  ByVal wParam As Long, _
  ByVal lParam As Long _
) As Long
    Dim I       As Long
    Dim HookObj As CHook    ' this instance is needed to raise the events
    
    For I = 0 To UBound(Hooks)
        ' found corresponding structure to current hWnd
        If Hooks(I).hWnd = hWnd Then _
            Exit For
    Next I
    
    If I > UBound(Hooks) Then
        ' not hooked
        ' aah, this should NEVER happen!
        Err.Raise -1, , "Window Procedure wrongly redirected"
    Else
        With Hooks(I)
            For I = 0 To UBound(.Messages)
                ' current message is indeed hooked!
                If .Messages(I) = uMsg Then _
                    Exit For
            Next I
            
            If I > UBound(.Messages) Then
                ' current message is not handled
                ' redirect to default window procedure
                WndProc = CallWindowProc(.PrevProc, .hWnd, uMsg, wParam, lParam)
            Else
                ' current message is hooked
                
                ' first create CHook instance from pointer
                Set HookObj = PtrToHook(.Handle)
                ' then raise event Before()
                HookObj.RaiseBefore uMsg, wParam, lParam
                
                If CBool(uMsg) Then
                    ' uMsg<>0, that means the user has not changed anything
                    ' redirect to defaul window procedure...
                    WndProc = CallWindowProc(.PrevProc, .hWnd, uMsg, wParam, lParam)
                    ' ...and raise the event After()
                    HookObj.RaiseAfter uMsg, wParam, lParam
                Else
                    ' if not so, the message flow has been deliberately interrupted
                    WndProc = 0
                End If ' CBool(uMsg)
                
                ' release CHook instance
                Set HookObj = Nothing
            End If ' I > UBound(.Messages)
        End With ' Hooks(I)
    End If ' I > UBound(Hooks)
End Function
'----------------------------------------------------

'----------------------------------------------------
' PtrToHook As CHook
' retrieves a CHook instance from a pointer
' to do this in an extra procedure means to reduce crashes:
' if the user halts the execution while in the window proc
' nothing happens
' do NEVER halt while in this procedure: program WILL crash!
' thanks to http://vbaccelerator.com for this trick
' - IN lpHook As Long: pointer to the CHook instance
' returns instance to the CHook class
'----------------------------------------------------
Private Function PtrToHook(lpHook As Long) As CHook
    Dim tmpObj As CHook
    
    ' create 'illegal' instance
    Call CopyMemory(tmpObj, lpHook, 4)
    ' set 'legal' instance
    Set PtrToHook = tmpObj
    ' destroy 'illegal' instance
    Call CopyMemory(tmpObj, 0&, 4)
End Function
